import { CompilerOptions, Node, TypeChecker } from 'typescript';
import { Resolver } from "./resolver/type";
export declare class MetadataGenerator {
    private readonly ignorePaths?;
    readonly nodes: Node[];
    readonly typeChecker: TypeChecker;
    private readonly program;
    private referenceTypes;
    private circularDependencyResolvers;
    private debugger;
    constructor(entryFile: string | Array<string>, compilerOptions: CompilerOptions, ignorePaths?: Array<string>);
    generate(): Metadata;
    isExportedNode(node: Node): boolean;
    addReferenceType(referenceType: Resolver.ReferenceType): void;
    getReferenceType(refName: string): Resolver.ReferenceType;
    onFinish(callback: (referenceTypes: {
        [refName: string]: Resolver.ReferenceType;
    }) => void): void;
    getClassDeclaration(className: string): Node;
    getInterfaceDeclaration(className: string): Node;
    private getSourceFiles;
    private buildControllers;
}
export interface Metadata {
    controllers: Array<Controller>;
    referenceTypes: {
        [typeName: string]: Resolver.ReferenceType;
    };
}
export interface Controller {
    location: string;
    methods: Array<Method>;
    name: string;
    path: string;
    consumes: Array<string>;
    produces: Array<string>;
    responses: Array<ResponseType>;
    tags: Array<string>;
    security?: Array<Security>;
}
export interface Method {
    deprecated?: boolean;
    description: string;
    method: string;
    name: string;
    parameters: Array<Parameter>;
    path: string;
    type: Resolver.BaseType;
    tags: Array<string>;
    responses: Array<ResponseType>;
    security?: Array<Security>;
    summary?: string;
    consumes: Array<string>;
    produces: Array<string>;
}
export interface Parameter {
    parameterName: string;
    description: string;
    in: string;
    name: string;
    required: boolean;
    type: Resolver.BaseType;
    collectionFormat?: boolean;
    allowEmptyValue?: boolean;
    default?: any;
    maxItems?: number;
    minItems?: number;
}
export interface Security {
    name: string;
    scopes?: Array<string>;
}
export interface ResponseType {
    description: string;
    status: string;
    schema?: Resolver.BaseType;
    examples?: any;
}
export interface Property {
    default?: any;
    format?: string;
    example?: unknown;
    validators: Record<string, {
        value?: any;
        message?: string;
    }>;
    description?: string;
    name: string;
    type: Resolver.Type;
    required: boolean;
}
export interface ResponseData {
    status: string;
    type: Resolver.BaseType;
}
