import { Property } from "../metadataGenerator";
export declare namespace Resolver {
    type TypeStringLiteral = 'string' | 'boolean' | 'double' | 'float' | 'file' | 'integer' | 'long' | 'enum' | 'array' | 'datetime' | 'date' | 'binary' | 'buffer' | 'byte' | 'void' | 'object' | 'any' | 'refEnum' | 'refObject' | 'refAlias' | 'nestedObjectLiteral' | 'union' | 'intersection';
    type RefTypeLiteral = 'refObject' | 'refEnum' | 'refAlias';
    type PrimitiveTypeLiteral = Exclude<TypeStringLiteral, RefTypeLiteral | 'enum' | 'array' | 'void' | 'nestedObjectLiteral' | 'union' | 'intersection'>;
    type Type = PrimitiveType | ObjectsNoPropsType | EnumType | ArrayType | FileType | DateTimeType | DateType | BinaryType | BufferType | ByteType | AnyType | RefEnumType | RefObjectType | RefAliasType | NestedObjectLiteralType | UnionType | IntersectionType;
    interface BaseType {
        typeName: TypeStringLiteral;
        typeArgument?: BaseType;
    }
    type PrimitiveType = StringType | BooleanType | DoubleType | FloatType | IntegerType | LongType | VoidType;
    interface StringType extends BaseType {
        typeName: 'string';
    }
    interface BooleanType extends BaseType {
        typeName: 'boolean';
    }
    interface DoubleType extends BaseType {
        typeName: 'double';
    }
    interface FloatType extends BaseType {
        typeName: 'float';
    }
    interface IntegerType extends BaseType {
        typeName: 'integer';
    }
    interface LongType extends BaseType {
        typeName: 'long';
    }
    interface VoidType extends BaseType {
        typeName: 'void';
    }
    function isVoidType(param: BaseType): param is VoidType;
    interface DateType extends BaseType {
        typeName: 'date';
    }
    interface FileType extends BaseType {
        typeName: 'file';
    }
    interface DateTimeType extends BaseType {
        typeName: 'datetime';
    }
    interface BinaryType extends BaseType {
        typeName: 'binary';
    }
    interface BufferType extends BaseType {
        typeName: 'buffer';
    }
    interface ByteType extends BaseType {
        typeName: 'byte';
    }
    interface AnyType extends BaseType {
        typeName: 'any';
    }
    interface ObjectsNoPropsType extends BaseType {
        typeName: 'object';
    }
    interface EnumType extends BaseType {
        members: Array<string | number | boolean | null>;
        typeName: 'enum';
    }
    function isEnumType(param: BaseType): param is EnumType;
    interface ArrayType extends BaseType {
        elementType: BaseType;
        typeName: 'array';
    }
    function isArrayType(param: BaseType): param is ArrayType;
    interface NestedObjectLiteralType extends BaseType {
        typeName: 'nestedObjectLiteral';
        properties: Array<Property>;
        additionalProperties?: Type;
    }
    function isNestedObjectLiteralType(param: BaseType): param is NestedObjectLiteralType;
    interface IntersectionType extends BaseType {
        typeName: 'intersection';
        members: Array<Type>;
    }
    function isIntersectionType(param: BaseType): param is IntersectionType;
    interface UnionType extends BaseType {
        typeName: 'union';
        members: Array<Type>;
    }
    function isUnionType(param: BaseType): param is UnionType;
    type ReferenceType = RefEnumType | RefObjectType | RefAliasType;
    interface ReferenceTypeBase extends BaseType {
        description?: string;
        typeName: RefTypeLiteral;
        refName: string;
        example?: unknown;
    }
    interface RefEnumType extends ReferenceTypeBase {
        typeName: 'refEnum';
        members: Array<string | number>;
        memberNames?: Array<string>;
    }
    function isRefEnumType(param: BaseType): param is RefEnumType;
    interface RefObjectType extends ReferenceTypeBase {
        typeName: 'refObject';
        properties: Array<Property>;
        additionalProperties?: Type;
    }
    interface RefAliasType extends Omit<Property, 'name' | 'required'>, ReferenceTypeBase {
        typeName: 'refAlias';
    }
    function isReferenceType(param: BaseType): param is ReferenceType;
}
export declare type PickUtilityType = 'Pick';
export declare type OmitUtilityType = 'Omit';
export declare type RequiredUtilityType = 'Required';
export declare type PartialUtilityType = 'Partial';
export declare type ReadonlyUtilityType = 'Readonly';
export declare type UtilityType = PickUtilityType | OmitUtilityType | PartialUtilityType | RequiredUtilityType | ReadonlyUtilityType;
